/* DCJavaClient - a DirectConnect client written in Java
 * Copyright (C) 2001 Stefan Grling, stefan@gorling.se
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

import java.io.*;
import java.net.*;
import java.util.*;
import java.math.*;
import DCEncryptionHandler;
import UserInfo;

/***

   This is the protocol-aware class. It handles connection traffic between two clients (i.e. downloads a file). 
   Everything protocol specific goes here.

 ***/

class DCClientProtocolHandler {
    Socket connection=null;
    GeneralBufferedReader inData=null;
    OutputStream outData=null;
    DCEncryptionHandler encryptionHandler=null;
    boolean hasRecievedLock=false;
    boolean debug=false;
    String remoteNick=null;
    boolean isLoggedIn=false;
    String remoteKey=null;
    UserInfo myself=null;
    
    final static char COMMAND_END_CHAR='|';
    final static char COMMAND_SEP_CHAR=' ' ;

    public DCClientProtocolHandler(Socket connection,UserInfo myself) throws IOException{
	if (debug)
	    System.out.println("DCClientProtocolHandler: Initiating Protocol Handler");
	this.connection=connection;
	this.myself=myself;
	if (debug)
	    System.out.println("DCClientProtocolHandler: Loading a general Buffered reader to handle incomming data.");

	// The GeneralBufferedReader is an extension of the BufferedReader, where the EOL-character may be manually specified.
	this.inData=new GeneralBufferedReader(new InputStreamReader(this.connection.getInputStream()),""+COMMAND_END_CHAR);

	if (debug)
	    System.out.println("DCClientProtocolHandler: Loading a output data stream");
	this.outData=new DataOutputStream(this.connection.getOutputStream());

	if (debug)
	    System.out.println("DCClientProtocolHandler: Initiating a DCEncryptionHandler");
	this.encryptionHandler=new DCEncryptionHandler();

	if (debug)
	    System.out.println("DCClientProtocolHandler: Done initiating Protocol Handler");

	if (debug)
	    System.out.println("DCClientProtocolHandler: Greeting remote host");

	// Say hello!
	sendCommand("$MyNick",myself.getNick());

	// Lock should be a randomized string, but i don't se the point.. It only verifies that we know the algorithm. 
	sendCommand("$Lock","AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA Pk=BBBBBBBBBBB");

    }


    public String getRemoteNick() throws IOException  {
	return waitForCommand("$MyNick",5000);
    }

    public void retrieveFile(SearchResult file) throws IOException {
	long startByte=0;

	while (!hasRecievedLock) {  // If the other end haven't introduced itself, wait for it, we'll need the key.
	    processBuffer();
	}

	String filename=file.filename.substring(file.filename.lastIndexOf("\\")+1,file.filename.length());

	if (debug)
	System.out.println("Saving file as :"+filename);
	//By here we have the $Lock from the other end, and we're ready to start speaking..

	sendCommand("$Direction","Download 9413");  // I want a download session, I have really no idea what 9413 means, it might be packet size, it might be size of the developers pants. We don't seem to need it anohow.
	sendCommand("$Key",remoteKey); // Give the client the response so that it's happy.

	if (debug)
		System.out.println("Asking for file:"+file.filename );

	if (debug && filename==null)
	    System.out.println("DCDownload - filename was null");

	File currentFile=new File(filename);
	RandomAccessFile infile=null;

	if (currentFile==null && debug) 
	    System.out.println("DCDownload currentFile was null");

	if (!currentFile.exists() && debug)
	    System.out.println("DCDownload - the current file didn't exist");
	if (!currentFile.isFile() && debug)
	    System.out.println("DCDownload - the current file wasn't file");

	if (!currentFile.canRead() && debug)
	    System.out.println("DCDownload - the current file couldn't be read");

	if (currentFile.exists() && currentFile.isFile() && currentFile.canRead() && file.mayResume) {
	    if (debug)
		System.out.println("DCDownload: found a file with the same name");
	    if (currentFile.length()<file.fileSize) {
		if (debug)
		    System.out.println("DCDownload: the file was smaller, trying to resume");

		if (debug)
		    System.out.println("DCDownload: file length="+currentFile.length());

		
		infile=new RandomAccessFile(currentFile,"rw"); // read the last 10 bytes so that we can verify that this is the same file.

		startByte=infile.length()-20; // move to 10 bytes from the end of the file.
		infile.seek(startByte); 
		
		if (debug)
		    System.out.println("DCDownload: tried to move "+startByte);
		
	    }else{
		filename=filename+".anotherFile"; //there was a problem reading the file, simply save it in another name.
		if (debug)
		    System.out.println("DCDownload: renaming file to:"+filename);
	    }
	}

	startByte++;
	sendCommand("$Get",file.filename+"$"+startByte);	
	int fileLength=Integer.parseInt(waitForCommand("$FileLength",20000));
	sendCommand("$Send","");

	file.fileSize=fileLength+startByte;
	fileLength-=startByte-1; // DC thinks a files first byte is at pos 1.

        if (startByte>1) { // We're resuming, loop over the last bytes to see that it is the same file.
	    int lastByte=1;
	    int i=0; // debug counter only!
	    boolean concistencyFailed=false; // misspelled?
	    while(lastByte!=-1) {
		lastByte=infile.read();
		if (lastByte!=-1) {
		    fileLength--;			
		    file.bytesRecieved++;
		    char c1=(char)lastByte;
		    char c2=(char)inData.read();
		    if (c1!=c2) {
			if (debug)
			    System.out.println(i+":DCDownload: ResumeConsistencyCheckFailed! - got "+c2+"("+(int)c2+") wanted:"+c1+" ("+(int)c1+")");
 
			concistencyFailed=true;
		    }else{
			if (debug)
			    System.out.println(i+":DCDownload- ResumeConsistencyCheck matched!");
		    }
		}
		i++;
	    }


	if (concistencyFailed) {
 	      file.mayResume=false; // we couldn't resume this one!
   	      throw new IOException();
	}

	}


	if (debug)
		System.out.println("Recieving data - FileLength:"+fileLength);

	file.transferInProgress=true;


        FileWriter fw=new FileWriter(filename,true); // append if available.
	int lastByte=1;
	while(fileLength>0 && lastByte!=-1) {
	  lastByte=inData.read();
	  if (lastByte!=-1) {
	    fw.write((char)lastByte);
   	    fw.flush();
            fileLength--;			
	    file.bytesRecieved++;
	  }
	}
	fw.flush();
	fw.close();

	if (lastByte==-1 && fileLength>0) { // if we got -1 - we have recieved a end of stream!
	      if (debug)
		System.out.println("DCDownloader: End of data stream reached, before end of download!");
	      throw new IOException();
	}


	file.transferFinnished=true;

	if (debug)
	    System.out.print("DONE");

    }


    public String waitForCommand(String command) throws IOException{
	return waitForCommand(command,30000);
    }

    public String waitForCommand(String command,int timeout) throws IOException{
	long started=System.currentTimeMillis();    

	while(true) { /// here we should specify some time of timeout.

	    if ((started+timeout)<System.currentTimeMillis()) {
		if (debug)
		System.out.println("TIMEOUT REACHED! Started"+started+" Timeout:"+timeout+" current:"+System.currentTimeMillis());
		throw new IOException();
	    }
   	StringTokenizer cmdTokens  = new StringTokenizer(inData.readLine(),""+COMMAND_END_CHAR);
   
	if (cmdTokens.hasMoreTokens()) {
	
	    while (cmdTokens.hasMoreTokens()) {
		String cmdString=cmdTokens.nextToken();

		if (debug)
		    System.out.println("Recieved:"+cmdString);

		    if(cmdString.equals("$MaxedOut")) {
			throw new IOException();
		    }


		if (cmdString.length()>1 && cmdString.indexOf(COMMAND_SEP_CHAR+"")>0 && (cmdString.indexOf(COMMAND_SEP_CHAR+"")+1)<(cmdString.length()-1)) { // Do not process empty commands
		    if (cmdString.indexOf(COMMAND_SEP_CHAR+"")>0) {

			// I wonder what $MaxedOut means.. - no slots perhaps?

	       	// Here we have recieved a command, a command has a "Command" anda a "Data" part, separated by a COMMAND_SEP_CHAR.
 		    String cmdStart=cmdString.substring(0,cmdString.indexOf(COMMAND_SEP_CHAR+""));
		    String cmdData=cmdString.substring(cmdString.indexOf(COMMAND_SEP_CHAR+"")+1,cmdString.length());
		    if (debug)
			System.out.println("Command was:#"+command+"Start:#"+cmdStart+"#");

		    if (command.equals(cmdStart)) {
			if (debug)
			    System.out.println("Returning cmdData");
			return cmdData;
		    }
		    	if (debug)
			    System.out.println("waiting for next");

			     
		    }
		}
	    }
	}
	}
    }

     

    public void processBuffer() throws IOException{
	    processCommand(inData.readLine());
    }


    /***
	processCommand takes a String with one or multiple commands, parses them and takes aproperate actions.
     ***/


    private void processCommand(String command) throws IOException{

   	StringTokenizer cmdTokens  = new StringTokenizer(command,""+COMMAND_END_CHAR);
   
	if (cmdTokens.hasMoreTokens()) {
	
	    while (cmdTokens.hasMoreTokens()) {
		String cmdString=cmdTokens.nextToken();
		
		if (cmdString.length()>1 && cmdString.indexOf(COMMAND_SEP_CHAR+"")>0 && (cmdString.indexOf(COMMAND_SEP_CHAR+"")+1)<(cmdString.length()-1)) { // Do not process empty commands
		    if (cmdString.indexOf(COMMAND_SEP_CHAR+"")>0) {
	       	// Here we have recieved a command, a command has a "Command" anda a "Data" part, separated by a COMMAND_SEP_CHAR.
 		    String cmdStart=cmdString.substring(0,cmdString.indexOf(COMMAND_SEP_CHAR+""));
		    String cmdData=cmdString.substring(cmdString.indexOf(COMMAND_SEP_CHAR+"")+1,cmdString.length());
		    if ("$Lock".equals(cmdStart)) {
			// This is the login Challenge/Reponse request, the challenge is the part before " Pk=".
			// I believe that Pk stands for privateKey or publicKey - However I haven't found any use for it
                        String cmdKey=cmdData.substring(0,cmdString.indexOf(" Pk=")-6);
       			String privateKey=cmdData.substring(cmdData.indexOf(" Pk=")+4,cmdData.length());	      					

			if (debug)
			    System.out.println("Recieved Challenge/Response request recieved, we're Logging in");			
			remoteKey=encryptionHandler.calculateValidationKey(cmdKey);
		       hasRecievedLock=true;
		    }else if ("$Direction".equals(cmdStart)) {
			System.out.println("Remote host tried to set upload direction:"+cmdData);
		    }else if ("$Key".equals(cmdStart)) {
			System.out.println("Remote host did send us a key, as if we cared");
		    }else if ("$FileLength".equals(cmdStart)) {

			System.out.println("Remote host gave us length:"+cmdData+", asking to send");
			sendCommand("$Send","");

			// After send, the data comes..

		    }else{
			if (debug)
			    System.out.println("Command unrecognized:"+cmdStart+" "+cmdData);
		    }
		    }
		}
	    }
	}
    }
		    
		    
    public boolean isLoggedIn() {
	return this.isLoggedIn;
    }


    public boolean logout() {
	try {
	connection.close();
	return true;
	}catch(Exception e) {
	    return false;
	}
    }


    private void sendCommand(String command,String data) throws IOException{
        outData.write((command+COMMAND_SEP_CHAR+data+COMMAND_END_CHAR).getBytes());
        outData.flush();
    }


}






