/* DCJavaClient - a DirectConnect client written in Java
 * Copyright (C) 2001 Stefan Grling, stefan@gorling.se
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


import java.io.*;
import java.net.*;
import java.util.*;
import java.math.*;
import DCEncryptionHandler;
import UserHandler;
import DownloadManager;
import DownloadManagerThread;
/***

   This is the protocol-aware class. It handles connection traffic. Everything protocol specific goes here.

 ***/

class DCProtocolHandler {
    Socket connection=null;
    GeneralBufferedReader inData=null;
    OutputStream outData=null;
    DCEncryptionHandler encryptionHandler=null;
    boolean isLoggedIn=false;
    boolean debug=false;

    
    UserInfo myself=null;

    UserHandler userHandler=new UserHandler(this);
    DownloadManager downloadManager=null;
    DownloadManagerThread downloadManagerThread=null;
    final static char COMMAND_END_CHAR='|';
    final static char COMMAND_SEP_CHAR=' ' ;

    public DCProtocolHandler(Socket connection,UserInfo myself) throws IOException{
	if (debug)
	System.out.println("DCProtocolHandler: Initiating Protocol Handler");

	this.connection=connection;
	this.myself=myself; 

	downloadManagerThread=new DownloadManagerThread(this);
	downloadManagerThread.start();
	downloadManager=downloadManagerThread.getDownloadManager();

	if (debug)
	System.out.println("DCProtocolHandler: Loading a general Buffered reader to handle incomming data.");

	// The GeneralBufferedReader is an extension of the BufferedReader, where the EOL-character may be manually specified.
	this.inData=new GeneralBufferedReader(new InputStreamReader(this.connection.getInputStream()),""+COMMAND_END_CHAR);

	if (debug)
	System.out.println("DCProtocolHandler: Loading a output data stream");
	this.outData=new DataOutputStream(this.connection.getOutputStream());

	if (debug)
	System.out.println("DCProtocolHandler: Initiating a DCEncryptionHandler");
	this.encryptionHandler=new DCEncryptionHandler();
	if (debug)
	System.out.println("DCProtocolHandler: Done initiating Protocol Handler");
    }

    public void processBuffer() throws IOException{
	    processCommand(inData.readLine());
    }


    /***
	processCommand takes a String with one or multiple commands, parses them and takes aproperate actions.
     ***/

    private void processCommand(String command) throws IOException{


   	StringTokenizer cmdTokens  = new StringTokenizer(command,""+COMMAND_END_CHAR);
   
	if (cmdTokens.hasMoreTokens()) {
	
	    while (cmdTokens.hasMoreTokens()) {
		String cmdString=cmdTokens.nextToken();

		
		if (cmdString.length()>1 && cmdString.indexOf(COMMAND_SEP_CHAR+"")>0 && (cmdString.indexOf(COMMAND_SEP_CHAR+"")+1)<(cmdString.length()-1)) { // Do not process empty commands
		    if (cmdString.indexOf(COMMAND_SEP_CHAR+"")>0) {


	       	// Here we have recieved a command, a command has a "Command" anda a "Data" part, separated by a COMMAND_SEP_CHAR.
 		    String cmdStart=cmdString.substring(0,cmdString.indexOf(COMMAND_SEP_CHAR+""));
		    String cmdData=cmdString.substring(cmdString.indexOf(COMMAND_SEP_CHAR+"")+1,cmdString.length());


		    if ("$Lock".equals(cmdStart)) {

			// This is the login Challenge/Reponse request, the challenge is the part before " Pk=".
			// I believe that Pk stands for privateKey or publicKey - However I haven't found any use for it

                        String cmdKey=cmdData.substring(0,cmdString.indexOf(" Pk=")-6);
       			String privateKey=cmdData.substring(cmdData.indexOf(" Pk=")+4,cmdData.length());	      					
			System.out.println("Recieved Challenge/Response request recieved, we're Logging in");
			
			String key=encryptionHandler.calculateValidationKey(cmdKey);
			sendCommand("$Key",key);
			sendCommand("$ValidateNick",myself.getNick());
			sendCommand("$GetNickList","");

			isLoggedIn=true;
		    }else if ("$HubName".equals(cmdStart)) {
			if (debug)
			    System.out.println("HubName is:"+cmdData);
		    }else if ("$ForceMove".equals(cmdStart)) {
			System.out.println("You are asked to redirect to:"+cmdData);
		    }else if (cmdStart.startsWith("<")) {
			System.out.println(cmdStart+cmdData);
		    }else if ("$RevConnectToMe".equals(cmdStart)) {
			if (debug)
			    System.out.println("Got revconnectome:"+cmdData);
		    }else if ("$Hello".equals(cmdStart)) {
			if (debug)
			    System.out.println("User arrived:#"+cmdData+"#");
			
			if (cmdData.equals(myself.getNick())) {
			    sendCommand("$MyINFO",myself.toString());
			    
			}

			// If we want a list of the users connected send a $GetNickList it returns a $$ separated nick-list 
			// We could send a $Version 1.000009 command (or larger version) but there's really no use. 

			// There is a connection request called $ConnectToMe myNick dest and a $RevConnectToMe remoteNick myNick - perhaps we should try that one someday :) - implemented

			// To search: $Search Hub:<MyNick> F?F?0?1?Office|
			// I suppose the F?F0?1? is a advanced-search pattern, tcpdump and advanced-parameter searches should give the ansver.

	       	    }else if ("$Quit".equals(cmdStart)) {
			if (debug)
			    System.out.println("User left"+cmdData);
			userHandler.removeUser(cmdData);
		    }else if ("$Search".equals(cmdStart)) {
			if (debug)
			    System.out.println("User searched:"+cmdData);
		    }else if ("$ConnectToMe".equals(cmdStart)) {
			if (debug)
			    System.out.println("User requested Connect:"+cmdData);
			
			downloadManager.enqueueConnectionRequest(cmdData.substring(cmdData.indexOf(" ")+1,cmdData.length()));
			
		    }else if ("$NickList".equals(cmdStart)) {
			if (debug)
			    System.out.println("gotNickList:"+cmdData);
			userHandler.addUsersFromNickList(cmdData);

		    }else if ("$OpList".equals(cmdStart)) {
			if (debug)
			    System.out.println("gotOpList:"+cmdData);
			
		    }else if ("$SR".equals(cmdStart)) { // SearchResult
			if (debug)
			    System.out.println("gotOpList:"+cmdData);
			downloadManager.addSearchResult(cmdData);
		    }else if ("$MyINFO".equals(cmdStart)) {
			if (debug)
			    System.out.println("User announced info:"+cmdData);
     	        	userHandler.addUser(cmdData);

		    }else{
			System.out.println("Command unrecognized:"+cmdStart+" "+cmdData);
		    }
		    }
		}
	    }
	}
    }
		    
		    
    public void makeSearch(String keyword) throws IOException {
	/* build a string like "$Search Hub:nick a?b?c?d?eeeee| where */
	/* a is F is size doesn't matter, else T */
	/* b is F if size is "at least", else T (at most) */
	/* c is the size in byte */
	/* d is data type: 1=any,2=audio,3=compressed,4=document,5=exe,6=picture,7=videos,8=folder */
	/* and eeee is the pattern to find */

	sendCommand("$Search","Hub:"+myself.getNick()+" F?F?0?1?"+keyword);
	downloadManager.clearSearchResults();
    }

    public void makeSearch(String keyword, int searchType, long minSize) throws IOException{
	sendCommand("$Search","Hub:"+myself.getNick()+" T?F?"+minSize+"?"+searchType+"?"+keyword);
	downloadManager.clearSearchResults();
    }

    public void getInfo(String nick) throws IOException {
	sendCommand("$GetINFO",nick+" "+myself.getNick());
    }

    public void requestConnection(String user) throws IOException {
	sendCommand("$RevConnectToMe",myself.getNick()+" "+user);
    }

    public void sendChatMessage(String message) throws IOException {
	sendCommand("<"+myself.getNick()+">",message);
    }

    public void sendPrivateMessage(String message,String nick) throws IOException {
	sendCommand("$To:",nick+" From: "+myself.getNick()+" $<"+myself.getNick()+">"+message);
    }

    public boolean isLoggedIn() {
	return this.isLoggedIn;
    }


    public boolean logout() {
	try {
	sendCommand("$Quit","");
	connection.close();
	return true;
	}catch(Exception e) {
	    return false;
	}
    }


    public UserInfo getMyInfo() {
	return myself;
    }
	private void sendCommand(String command,String data) throws IOException{
	    if (debug)
		System.out.println("Sending command:"+command+COMMAND_SEP_CHAR+data+COMMAND_END_CHAR+COMMAND_END_CHAR);

	    outData.write((command+COMMAND_SEP_CHAR+data+COMMAND_END_CHAR).getBytes());
	    outData.flush();
	}


    public DownloadManager getDownloadManager() {
	return downloadManager;
    }

    public UserHandler getUserHandler() {
	return userHandler;
    }

}






